#include "shmem.h"
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <signal.h>


/* paranoid checdk */
#define PARANOID


/* primitives */
static int shm_create(int size)
{
	/* maximum number of tries to get shared memory handle */
	int maxtries = 1000;

	/* keep on trying random keys */
	while (maxtries--) {
                          int shmid = shmget( rand(), size, IPC_CREAT | IPC_EXCL|0666);
		if (-1 != shmid  ) {
			printf("shmem.shm_create() - shmid = %d\n",shmid);
			return shmid;
			}
		}

	fprintf(stderr,"File %s, at line %d, shmem.shm_create() - unable to get a handle for a shared memory block of size %d, aborting !\n",
		__FILE__, __LINE__, size);
	return -1;
}

int shm_create_pool(shmem *self, int i)
{
	char *ptr;

	/* allocate shared memory and get a handle */
	self->pools[i] = shm_create(MEMPOOL_SIZE);
	if (self->pools[i] == -1)
		return -1;

	/* use a pointer temporarily to format */
	ptr = shmat(self->pools[i], 0, 0);
	self->m[i] = mempool_create(ptr,MEMPOOL_SIZE);	

	return 0;
}

char *shm_malloc(mempool *m, int size)
{
	return mempool_malloc(m,size);
}


/**************************************************************************************/



/*
 * shmem API
 */

int shmem_aquire(shmem *self, long timeout)
{
	int elapsed = 0;
	int maxtries = 1000;
	int i;

#ifdef PARANOID
printf("entering shmem_aquire\n");
#endif

	/* first try to lock memory */
	while ((maxtries--) && self->inuse) {
		usleep( LOCK_POLL_RATE );
		elapsed += LOCK_POLL_RATE;
		if (elapsed > timeout)
			return -1; 		
		}
	self->inuse = 1;

	/* now map handles to shared memory pools to pointers */
	for (i = 0; i < self->top; i++)
		self->m[i] = shmat(self->pools[i],0,0);

#ifdef PARANOID
printf("leaving shmem_aquire\n");
#endif


	return 0;	
}


void shmem_release(shmem *self)
{
	int i;
	
#ifdef PARANOID
printf("entering shmem_release\n");
#endif


	for (i = 0; i < self->top; i++)
		shmdt(self->m[i]);
	self->inuse = 0;	

#ifdef PARANOID
printf("leaving shmem_release\n");
#endif

}


shmem *shmem_open(int shmid)
{
	return shmat(shmid,0,0);
}

void shmem_close(shmem *self)
{
	shmdt(self);
}


shmem *shmem_create()
{
	shmem *s;
	int shmid, i;


#ifdef PARANOID
printf("entering shmem_create\n");
#endif

	/* allocate shmem in shared memory */
	shmid = shm_create(sizeof(shmem));
	if (shmid == -1)
		return NULL;

	/* map handle to a pointer in my namespace */
	s = shmat(shmid,0,0);	

	/* initialize structure */
	s->shmid = shmid;
	s->inuse = 0;
	s->top = 0;
	for (i = 0; i < MAX_NUM_POOLS; i++)
		s->pools[i] = -1;
#ifdef PARANOID
printf("leaving shmem_create\n");
#endif

	return s;
}

void shmem_delete(shmem *self)
{
	register int i;
#ifdef PARANOID
printf("entering shmem_delete\n");
#endif

	/* mark all pools for deletion*/
	for (i = 0; i < self->top; i++)
		shmctl(self->pools[i],IPC_RMID,0);

	/* mark the shmem for deletion */
	shmctl( self->shmid,IPC_RMID,0 );

	/* remove pointer reference */
	shmdt( self );
#ifdef PARANOID
printf("leaving shmem_delete\n");
#endif
}

int shmem_size(shmem *self, char *ptr)
{
	mp_block_header *h;

#ifdef PARANOID
printf("entering shmem_size\n");
#endif
	/* go back and get the header */
	/* |mempool header|mempool number| data		*/
	((char *) h) = ptr - sizeof(int) - sizeof(mp_block_header);
#ifdef PARANOID
printf("leaving shmem_size -> %d\n", h->size);
#endif

	return h->size;
}

void shmem_free(shmem *self, char *ptr)
{
	mempool *m;
	int index;

#ifdef PARANOID
printf("entering shmem_free\n");
#endif
	/* extract the pool number and get the shared memory id */
	memcpy( &index, (ptr - sizeof(int)), sizeof(int) );
	m = self->m[index];

	/* free memory */
	mempool_free(m, (ptr-sizeof(int)) );

#ifdef PARANOID
printf("leaving shmem_free\n");
#endif
	return;	
}



char *shmem_calloc(shmem *self, int size)
{
	char *ptr;
#ifdef PARANOID
printf("entering shmem_calloc\n");
#endif
	ptr = shmem_malloc(self, size);
	memset(ptr,0,size);
#ifdef PARANOID
printf("leaving shmem_calloc\n");
#endif
	return ptr;
}

char *shmem_realloc(shmem *self, char *old, int nsize)
{
	char *ptr;
	int osize;


#ifdef PARANOID
printf("entering shmem_realloc\n");
#endif
	/* if null then behave like malloc */
	if (old == NULL) {
                           return shmem_malloc(self,nsize);
		}

	osize = shmem_size(self,old);
	
	if ((ptr = shmem_malloc(self, nsize)) == NULL)
		return NULL;

	/* fill new memory */
	if (osize > nsize)	
		memcpy(ptr,old,nsize);
	else
		memcpy(ptr,old,osize);

	/* free old */
	shmem_free(self, old);
	
#ifdef PARANOID
printf("leaving shmem_realloc\n");
#endif
	return ptr;
}



char *shmem_malloc(shmem *self, int size)
{
	char *ptr;
	int i;

#ifdef PARANOID
printf("entering shmem_malloc\n");
#endif

	/* increase size to accomodate embedded pool number*/
	size += sizeof(int);

	if (size > MEMPOOL_SIZE) {
                          fprintf(stderr,"File %s, at line %d: shmem.shmem_malloc() size -> %d  is greater than MAXPOOL -> %d\n",
			__FILE__, __LINE__, size, MEMPOOL_SIZE);
		return NULL;
		}

	for (i = 0; i < self->top; i++) {
		if (self->pools[i] != -1) { /* paranoid check */
			ptr = shm_malloc(self->m[i], size);
			if ( ptr )  {
				/* embedd pool number */
		                        	memcpy( ptr, &i, sizeof(int));
#ifdef PARANOID
printf("leaving shmem_malloc\n");
#endif
				return &ptr[sizeof(int)];
				}			
			}
		}

	/* bounds check */
	if ((self->top+1) ==  MAX_NUM_POOLS) {
                          fprintf(stderr,"File %s, at line %d: shmem.shmem_malloc() - memory exhausted !! \n",
			__FILE__, __LINE__ );
		return NULL;
		}

	/* create a new pool */
	if (shm_create_pool(self, self->top))
		return NULL;

	/* now try and allocate from this new pool */
	ptr = shm_malloc(self->m[self->top], size);		
	if (ptr == NULL)
		return NULL;

	/* embedd pool number at the head */
       	memcpy( ptr, &i, sizeof(int));

	/* increment top of stack */
	self->top++;

#ifdef PARANOID
printf("leaving shmem_malloc\n");
#endif

	return &ptr[sizeof(int)];
}


#ifdef TEST_SHMEM

void test2_owner()
{
	shmem *s;

	s = shmem_create();

	printf("owner shmid = %d\n",s->shmid);

	while (s->inuse == 0);

	printf("Memory in use !!\n");

	while (s->inuse == 1);

	shmem_aquire( s, 1000 );

	mempool_dump( s->m[0] );

	shmem_release( s );

	shmem_delete( s );	
}

void test2_user(int shmid)
{
	shmem *s;
	char *ptr;

	s = shmem_open(shmid);
	shmem_aquire( s, 1000 );


	usleep( 1000 );

	if ((ptr = shmem_malloc(s,10)) == NULL) {
		printf("shmem_malloc failed\n");
		} else {

		sprintf(ptr,"Hello");

		}
		
	shmem_release( s );
	shmem_close( s );
}


void test1()
{
	shmem *s;
	char *ptr;

	s = shmem_create();
	if (shmem_aquire(s, 10000)) {
		printf("Timout !!\n");
		return;
		}

	if ((ptr = shmem_malloc(s,100)) == NULL) {
		printf("shmem_malloc failed\n");
		return;
		}
	shmem_free(s,ptr);

	shmem_release(s);

	shmem_delete( s );	
}



int main(int argc, char *argv[])
{
	switch( atoi(argv[1]) )
	{
             case 0:
		test1();
		break;
	case 1:
		test2_owner();
		break;
	case 2:
		test2_user( atoi(argv[2]) );
		break;
	}

	return 0;
}



#endif








